@@ -1,4 +1,4 @@
-/* Copyright (C) 2001, 2002, 2003 Red Hat, Inc.
+/* Copyright (C) 2001, 2002, 2003, 2004 Red Hat, Inc.
    Written by Jakub Jelinek <jakub@redhat.com>, 2001.
 
    This program is free software; you can redistribute it and/or modify
@@ -20,8 +20,10 @@
 #include <errno.h>
 #include <error.h>
 #include <fcntl.h>
+#include <fnmatch.h>
 #include <ftw.h>
 #include <glob.h>
+#include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -46,6 +48,13 @@ static struct prelink_dir *blacklist;
 static char *blacklist_dir;
 static size_t blacklist_dir_len;
 #endif
+static struct extension
+{
+  const char *ext;
+  size_t len;
+  int is_glob;
+} *blacklist_ext;
+static int blacklist_next;
 
 static int
 gather_deps (DSO *dso, struct prelink_entry *ent)
@@ -66,7 +75,11 @@ gather_deps (DSO *dso, struct prelink_entry *ent)
   const char *dl;
 
   if (check_dso (dso))
+    {
+      if (! undo)
+	ent->type = ET_UNPRELINKABLE;
       goto error_out;
+    }
 
   ent->pltgot = dso->info[DT_PLTGOT];
   ent->soname = strdup (dso->soname);
@@ -95,9 +108,10 @@ gather_deps (DSO *dso, struct prelink_entry *ent)
 
   for (i = 1; i < dso->ehdr.e_shnum; ++i)
     {
-      const char *name
-	= strptr (dso, dso->ehdr.e_shstrndx, dso->shdr[i].sh_name);
-      if (! strcmp (name, ".gnu.liblist")
+      const char *name;
+      if (dso->shdr[i].sh_type == SHT_GNU_LIBLIST
+	  && (name = strptr (dso, dso->ehdr.e_shstrndx, dso->shdr[i].sh_name))
+	  && ! strcmp (name, ".gnu.liblist")
 	  && (dso->shdr[i].sh_size % sizeof (Elf32_Lib)) == 0)
 	{
 	  nliblist = dso->shdr[i].sh_size / sizeof (Elf32_Lib);
@@ -113,7 +127,11 @@ gather_deps (DSO *dso, struct prelink_entry *ent)
 	  if (! undo)
 	    break;
 	}
-      else if (undo && ! strcmp (name, ".gnu.prelink_undo"))
+      else if (undo
+	       && dso->shdr[i].sh_type == SHT_PROGBITS
+	       && (name = strptr (dso, dso->ehdr.e_shstrndx,
+				  dso->shdr[i].sh_name))
+	       && ! strcmp (name, ".gnu.prelink_undo"))
 	ent->done = 2;
     }
 
@@ -154,7 +172,7 @@ gather_deps (DSO *dso, struct prelink_entry *ent)
 	  q = strstr (p, " (");
 	  if (q == NULL && strcmp (p, " => not found") == 0)
 	    {
-	      error (0, 0, "%s: Could not find one of dependencies",
+	      error (0, 0, "%s: Could not find one of the dependencies",
 		     ent->filename);
 	      goto error_out;
 	    }
@@ -164,7 +182,23 @@ gather_deps (DSO *dso, struct prelink_entry *ent)
 	  if (strstr (line, "statically linked") != NULL)
 	    error (0, 0, "%s: Library without dependencies", ent->filename);
 	  else
-	    error (0, 0, "%s: Could not parse `%s'", ent->filename, line);
+	    {
+	      p = strstr (line, "error while loading shared libraries: ");
+	      if (p != NULL)
+		{
+		  p += sizeof "error while loading shared libraries: " - 1;
+		  q = strstr (line, "cannot open shared object file: "
+				    "No such file or directory");
+		  if (q != NULL)
+		    {
+		      error (0, 0,
+			     "%s: Could not find one of the dependencies",
+			     ent->filename);
+		      goto error_out;
+		    }
+		}
+	      error (0, 0, "%s: Could not parse `%s'", ent->filename, line);
+	    }
 	  goto error_out;
 	}
 
@@ -227,11 +261,17 @@ gather_deps (DSO *dso, struct prelink_entry *ent)
     }
 
   ent->ndepends = ndepends;
+  char *cache_dyn_depends = NULL;
+  if (ndepends)
+    {
+      cache_dyn_depends = alloca (ndepends);
+      memset (cache_dyn_depends, '\0', ndepends);
+    }
   for (i = 0; i < ndepends; ++i)
     {
       ent->depends[i] = prelink_find_entry (depends [i], NULL, 1);
       if (ent->depends[i] == NULL)
-	goto error_out;
+	goto error_out_free_depends;
 
       if (ent->depends[i]->type == ET_CACHE_DYN)
 	{
@@ -239,14 +279,22 @@ gather_deps (DSO *dso, struct prelink_entry *ent)
 	  free (ent->depends[i]->depends);
 	  ent->depends[i]->depends = NULL;
 	  ent->depends[i]->ndepends = 0;
+	  cache_dyn_depends[i] = 1;
 	}
 
       if (ent->depends[i]->type != ET_NONE
 	  && ent->depends[i]->type != ET_BAD
-	  && ent->depends[i]->type != ET_DYN)
+	  && ent->depends[i]->type != ET_DYN
+	  && ent->depends[i]->type != ET_UNPRELINKABLE)
 	{
 	  error (0, 0, "%s is not a shared library", depends [i]);
-	  goto error_out;
+error_out_regather_libs:
+	  for (i = 0; i < ndepends; ++i)
+	    {
+	      if (cache_dyn_depends[i] && ent->depends[i]->type == ET_NONE)
+		gather_lib (ent->depends[i]);
+	    }
+	  goto error_out_free_depends;
 	}
     }
 
@@ -256,16 +304,28 @@ gather_deps (DSO *dso, struct prelink_entry *ent)
   for (i = 0; i < ndepends; ++i)
     if (ent->depends[i]->type == ET_NONE
 	&& gather_lib (ent->depends[i]))
-      goto error_out;
+      {
+	cache_dyn_depends[i] = 0;
+	goto error_out_regather_libs;
+      }
 
   for (i = 0; i < ndepends; ++i)
     for (j = 0; j < ent->depends[i]->ndepends; ++j)
       if (ent->depends[i]->depends[j] == ent)
 	{
-	  error (0, 0, "%s has dependency cycle", ent->canon_filename);
-	  goto error_out;
+	  error (0, 0, "%s has a dependency cycle", ent->canon_filename);
+	  goto error_out_free_depends;
 	}
 
+  for (i = 0; i < ndepends; ++i)
+    if (ent->depends[i]->type == ET_UNPRELINKABLE)
+      {
+	error (0, 0, "Could not prelink %s because its dependency %s could not be prelinked",
+	       ent->filename, ent->depends[i]->filename);
+	ent->type = ET_UNPRELINKABLE;
+	goto error_out;
+      }
+
   if (! undo && (!nliblist || liblist) && nliblist == ndepends)
     {
       for (i = 0; i < ndepends; ++i)
@@ -280,13 +340,14 @@ gather_deps (DSO *dso, struct prelink_entry *ent)
 
   return 0;
 
+error_out_free_depends:
+  free (ent->depends);
+  ent->depends = NULL;
+  ent->ndepends = 0;
 error_out:
   if (f)
     execve_close (f);
   free (line);
-  free (ent->depends);
-  ent->depends = NULL;
-  ent->ndepends = 0;
   free (depends);
   if (dso)
     close_dso (dso);
@@ -372,7 +433,7 @@ gather_dso (DSO *dso, struct prelink_entry *ent)
 	  last = -1;
 	  for (i = 0; i < dso->ehdr.e_phnum; ++i)
 	    if (dso->phdr[i].p_type == PT_LOAD
-	        && dso->phdr[i].p_vaddr + dso->phdr[i].p_memsz >= start)
+		&& dso->phdr[i].p_vaddr + dso->phdr[i].p_memsz >= start)
 	      {
 		if (last != -1
 		    && (((dso->phdr[last].p_vaddr + dso->phdr[last].p_memsz
@@ -439,7 +500,21 @@ gather_exec (DSO *dso, const struct stat64 *st)
 
   /* If there are no PT_INTERP segments, it is statically linked.  */
   if (i == dso->ehdr.e_phnum)
-    goto error_out;
+    {
+make_unprelinkable:
+      if (undo)
+	goto error_out;
+
+      ent = prelink_find_entry (dso->filename, st, 1);
+      if (ent == NULL)
+	goto error_out;
+
+      assert (ent->type == ET_NONE);
+      ent->type = ET_UNPRELINKABLE;
+      if (dso)
+	close_dso (dso);
+      return 0;
+    }
 
   j = addr_to_sec (dso, dso->phdr[i].p_vaddr);
   if (j == -1 || dso->shdr[j].sh_addr != dso->phdr[i].p_vaddr
@@ -447,7 +522,7 @@ gather_exec (DSO *dso, const struct stat64 *st)
     {
       error (0, 0, "%s: PT_INTERP segment not corresponding to .interp section",
 	     dso->filename);
-      goto error_out;
+      goto make_unprelinkable;
     }
 
   data = elf_getdata (dso->scn[j], NULL);
@@ -552,7 +627,7 @@ static int
 gather_func (const char *name, const struct stat64 *st, int type,
 	     struct FTW *ftwp)
 {
-  unsigned char e_ident [EI_NIDENT + 2];
+  unsigned char e_ident [sizeof (Elf64_Ehdr) + sizeof (Elf64_Phdr)];
 
 #ifndef HAVE_FTW_ACTIONRETVAL
   if (blacklist_dir)
@@ -565,20 +640,48 @@ gather_func (const char *name, const struct stat64 *st, int type,
 #endif
   if (type == FTW_F && S_ISREG (st->st_mode) && (st->st_mode & 0111))
     {
-      int fd;
+      int fd, i;
       DSO *dso;
       struct prelink_entry *ent;
+      size_t len = strlen (name);
+      const char *base = NULL;
+
+      for (i = 0; i < blacklist_next; ++i)
+	if (blacklist_ext[i].is_glob)
+	  {
+	    if (base == NULL)
+	      {
+		base = strrchr (name, '/');
+		if (base == NULL)
+		  base = name;
+		else
+		  ++base;
+	      }
+	    if (fnmatch (blacklist_ext[i].ext, base, FNM_PERIOD) == 0)
+	      return FTW_CONTINUE;
+	  }
+	else if (blacklist_ext[i].len <= len
+		 && memcmp (name + len - blacklist_ext[i].len,
+			    blacklist_ext[i].ext, blacklist_ext[i].len) == 0)
+	  return FTW_CONTINUE;
 
       ent = prelink_find_entry (name, st, 0);
       if (ent != NULL && ent->type != ET_NONE)
 	{
-	  if (verbose > 5
-	      && (ent->type == ET_CACHE_EXEC || ent->type == ET_CACHE_DYN))
-	    printf ("Assuming prelinked %s\n", name);
+	  if (verbose > 5)
+	    {
+	      if (ent->type == ET_CACHE_EXEC || ent->type == ET_CACHE_DYN)
+		printf ("Assuming prelinked %s\n", name);
+	      if (ent->type == ET_UNPRELINKABLE)
+		printf ("Assuming non-prelinkable %s\n", name);
+	    }
 	  ent->u.explicit = 1;
 	  return FTW_CONTINUE;
 	}
 
+      if (st->st_size < sizeof (e_ident))
+	return FTW_CONTINUE;
+
       fd = open (name, O_RDONLY);
       if (fd == -1)
 	return FTW_CONTINUE;
@@ -590,23 +693,127 @@ close_it:
 	  return FTW_CONTINUE;
 	}
 
-      /* Quickly find ET_EXEC ELF binaries only.  */
+      /* Quickly find ET_EXEC ELF binaries and most of PIE binaries.  */
 
       if (memcmp (e_ident, ELFMAG, SELFMAG) != 0)
-	goto close_it;
+	{
+make_unprelinkable:
+	  if (! undo)
+	    {
+	      ent = prelink_find_entry (name, st, 1);
+	      if (ent != NULL)
+		{
+		  assert (ent->type == ET_NONE);
+		  ent->type = ET_UNPRELINKABLE;
+		}
+	    }
+	  close (fd);
+	  return FTW_CONTINUE;
+	}
 
       switch (e_ident [EI_DATA])
 	{
 	case ELFDATA2LSB:
-	  if (e_ident [EI_NIDENT] != ET_EXEC || e_ident [EI_NIDENT + 1] != 0)
-	    goto close_it;
+	  if (e_ident [EI_NIDENT + 1] != 0)
+	    goto make_unprelinkable;
+	  if (e_ident [EI_NIDENT] != ET_EXEC)
+	    {
+	      if (e_ident [EI_NIDENT] != ET_DYN)
+		goto make_unprelinkable;
+	      else if (e_ident [EI_CLASS] == ELFCLASS32)
+		{
+		  if (e_ident [offsetof (Elf32_Ehdr, e_phoff)]
+		      == sizeof (Elf32_Ehdr)
+		      && memcmp (e_ident + offsetof (Elf32_Ehdr, e_phoff) + 1,
+				 "\0\0\0", 3) == 0
+		      && (e_ident [offsetof (Elf32_Ehdr, e_phnum)]
+			  || e_ident [offsetof (Elf32_Ehdr, e_phnum) + 1])
+		      && e_ident [sizeof (Elf32_Ehdr)
+				  + offsetof (Elf32_Phdr, p_type)] == PT_PHDR
+		      && memcmp (e_ident + sizeof (Elf32_Ehdr)
+				 + offsetof (Elf32_Phdr, p_type) + 1,
+				 "\0\0\0", 3) == 0)
+		    {
+maybe_pie:
+		      dso = fdopen_dso (fd, name);
+		      if (dso == NULL)
+			goto close_it;
+		      if (dynamic_info_is_set (dso, DT_DEBUG))
+			{
+			  close_dso (dso);
+			  goto make_unprelinkable;
+			}
+		      close_dso (dso);
+		    }
+		  goto close_it;
+		}
+	      else if (e_ident [EI_CLASS] == ELFCLASS64)
+		{
+		  if (e_ident [offsetof (Elf64_Ehdr, e_phoff)]
+		      == sizeof (Elf64_Ehdr)
+		      && memcmp (e_ident + offsetof (Elf64_Ehdr, e_phoff) + 1,
+				 "\0\0\0\0\0\0\0", 7) == 0
+		      && (e_ident [offsetof (Elf64_Ehdr, e_phnum)]
+			  || e_ident [offsetof (Elf64_Ehdr, e_phnum) + 1])
+		      && e_ident [sizeof (Elf64_Ehdr)
+				  + offsetof (Elf64_Phdr, p_type)] == PT_PHDR
+		      && memcmp (e_ident + sizeof (Elf64_Ehdr)
+				 + offsetof (Elf64_Phdr, p_type) + 1,
+				 "\0\0\0", 3) == 0)
+		    goto maybe_pie;
+		  goto close_it;
+		}
+	      else
+		goto make_unprelinkable;
+	    }
 	  break;
 	case ELFDATA2MSB:
-	  if (e_ident [EI_NIDENT + 1] != ET_EXEC || e_ident [EI_NIDENT] != 0)
-	    goto close_it;
+	  if (e_ident [EI_NIDENT] != 0)
+	    goto make_unprelinkable;
+	  if (e_ident [EI_NIDENT + 1] != ET_EXEC)
+	    {
+	      if (e_ident [EI_NIDENT + 1] != ET_DYN)
+		goto make_unprelinkable;
+	      else if (e_ident [EI_CLASS] == ELFCLASS32)
+		{
+		  if (e_ident [offsetof (Elf32_Ehdr, e_phoff) + 3]
+		      == sizeof (Elf32_Ehdr)
+		      && memcmp (e_ident + offsetof (Elf32_Ehdr, e_phoff),
+				 "\0\0\0", 3) == 0
+		      && (e_ident [offsetof (Elf32_Ehdr, e_phnum)]
+			  || e_ident [offsetof (Elf32_Ehdr, e_phnum) + 1])
+		      && e_ident [sizeof (Elf32_Ehdr)
+				  + offsetof (Elf32_Phdr, p_type) + 3]
+			 == PT_PHDR
+		      && memcmp (e_ident + sizeof (Elf32_Ehdr)
+				 + offsetof (Elf32_Phdr, p_type),
+				 "\0\0\0", 3) == 0)
+		    goto maybe_pie;
+		  goto close_it;
+		}
+	      else if (e_ident [EI_CLASS] == ELFCLASS64)
+		{
+		  if (e_ident [offsetof (Elf64_Ehdr, e_phoff) + 7]
+		      == sizeof (Elf64_Ehdr)
+		      && memcmp (e_ident + offsetof (Elf64_Ehdr, e_phoff),
+				 "\0\0\0\0\0\0\0", 7) == 0
+		      && (e_ident [offsetof (Elf64_Ehdr, e_phnum)]
+			  || e_ident [offsetof (Elf64_Ehdr, e_phnum) + 1])
+		      && e_ident [sizeof (Elf64_Ehdr)
+				  + offsetof (Elf64_Phdr, p_type) + 3]
+			 == PT_PHDR
+		      && memcmp (e_ident + sizeof (Elf64_Ehdr)
+				 + offsetof (Elf64_Phdr, p_type),
+				 "\0\0\0", 3) == 0)
+		    goto maybe_pie;
+		  goto close_it;
+		}
+	      else
+		goto make_unprelinkable;
+	    }
 	  break;
 	default:
-	  goto close_it;
+	  goto make_unprelinkable;
 	}
 
       dso = fdopen_dso (fd, name);
@@ -622,7 +829,7 @@ close_it:
       default: return FTW_STOP;
       case 2:
 #ifdef HAVE_FTW_ACTIONRETVAL
-        return FTW_SKIP_SUBTREE;
+	return FTW_SKIP_SUBTREE;
 #else
 	{
 	  blacklist_dir_len = strlen (name) + 1;
@@ -660,6 +867,13 @@ gather_binlib (const char *name, const struct stat64 *st)
     }
 
   ent = prelink_find_entry (name, st, 0);
+  if (ent != NULL && ent->type == ET_UNPRELINKABLE)
+    {
+      free (ent->depends);
+      ent->depends = NULL;
+      ent->ndepends = 0;
+      ent->type = ET_NONE;
+    }
   if (ent != NULL && ent->type != ET_NONE)
     {
       ent->u.explicit = 1;
@@ -884,7 +1098,7 @@ static int
 gather_check_lib (void **p, void *info)
 {
   struct prelink_entry *e = * (struct prelink_entry **) p;
-    
+
   if (e->type != ET_DYN)
     return 1;
 
@@ -913,7 +1127,7 @@ gather_check_lib (void **p, void *info)
 
       if (dir != NULL)
 	{
-	  error (0, 0, "%s is presnet in a blacklisted directory %s",
+	  error (0, 0, "%s is present in a blacklisted directory %s",
 		 e->canon_filename, dir->dir);
 	  e->type = ET_BAD;
 	  return 1;
@@ -1031,6 +1245,27 @@ add_to_blacklist (const char *name, int deref, int onefs)
   return 0;
 }
 
+void
+add_blacklist_ext (const char *ext)
+{
+  blacklist_ext = realloc (blacklist_ext,
+			   (blacklist_next + 1) * sizeof (*blacklist_ext));
+  if (blacklist_ext == NULL)
+    error (EXIT_FAILURE, errno, "can't create blacklist extension list");
+  if (*ext == '*' && strpbrk (ext + 1, "*?[{") == NULL)
+    {
+      blacklist_ext[blacklist_next].is_glob = 0;
+      ext++;
+    }
+  else
+    blacklist_ext[blacklist_next].is_glob = 1;
+  blacklist_ext[blacklist_next].ext = strdup (ext);
+  if (blacklist_ext[blacklist_next].ext == NULL)
+    error (EXIT_FAILURE, errno, "can't create blacklist extension list");
+  blacklist_ext[blacklist_next].len = strlen (ext);
+  blacklist_next++;
+}
+
 int
 blacklist_from_config (const char *config)
 {
@@ -1080,6 +1315,12 @@ blacklist_from_config (const char *config)
       if (*p == '\0' || !blacklist)
 	continue;
 
+      if (strchr (p, '/') == NULL)
+	{
+	  add_blacklist_ext (p);
+	  continue;
+	}
+
       if (strpbrk (p, "*?[{") == NULL)
 	{
 	  ret = add_to_blacklist (p, deref, onefs);
@@ -1093,7 +1334,7 @@ blacklist_from_config (const char *config)
 	{
 	  glob_t g;
 
-	  if (!glob (p, GLOB_BRACE, NULL, &g))
+	  if (!glob (p, GLOB_BRACE | GLOB_PERIOD, NULL, &g))
 	    {
 	      size_t n;
 
